@@ -15,6 +15,7 @@
 #    You should have received a copy of the GNU General Public License
 #    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 import asyncio
+from base64 import b64encode
 import functools
 import logging
 import random
@@ -22,7 +23,7 @@ import re
 import ssl
 import sys
 import time
-from asyncio import BaseTransport, Future
+from asyncio import BaseTransport, Future, Task
 from collections import namedtuple
 from copy import copy
 from datetime import datetime, timezone, timedelta
@@ -139,7 +140,7 @@ def arguments_rfs2971(**kwargs: Union[dict, list, str]) -> Union[dict, list]:
     return args
 
 
-class Command(object):
+class Command:
     def __init__(self, name: str, tag: str, *args, prefix: str = None, untagged_resp_name: str = None,
                  loop: asyncio.AbstractEventLoop = None, timeout: float = None) -> None:
         self.name = name
@@ -155,7 +156,7 @@ class Command(object):
         self._timer = asyncio.Handle(lambda: None, None, self._loop)  # fake timer
         self._set_timer()
         self._expected_size = 0
-        
+
         self._resp_literal_data = bytearray()
         self._resp_result = 'Init'
         self._resp_lines: List[bytes] = list()
@@ -163,7 +164,8 @@ class Command(object):
     def __repr__(self) -> str:
         return '{tag} {prefix}{name}{space}{args}'.format(
             tag=self.tag, prefix=self.prefix or '', name=self.name,
-            space=' ' if self.args else '', args=' '.join(self.args))
+            space=' ' if self.args else '', args=' '.join(str(arg) if arg is not None else '' \
+                                                          for arg in self.args))
 
     # for tests
     def __eq__(self, other):
@@ -329,6 +331,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         self.incomplete_line = b''
         self.current_command = None
         self.conn_lost_cb = conn_lost_cb
+        self.tasks: set[Future] = set()
 
         self.tagnum = 0
         self.tagpre = int2ap(random.randint(4096, 65535))
@@ -388,7 +391,9 @@ class IMAP4ClientProtocol(asyncio.Protocol):
             return
 
         if self.state == CONNECTED:
-            asyncio.ensure_future(self.welcome(line))
+            task = asyncio.ensure_future(self.welcome(line))
+            self.tasks.add(task)
+            task.add_done_callback(self.tasks.discard)
         elif tagged_status_response_re.match(line):
             self._response_done(line)
         elif current_cmd is not None:
@@ -401,12 +406,15 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         else:
             log.info('unknown data received %s' % line)
 
-    def send(self, line: str) -> None:
+    def send(self, line: str, scrub: str =None) -> None:
         data = ('%s\r\n' % line).encode()
-        log.debug('Sending : %s' % data)
+        if scrub:
+            log.debug('Sending : %s' % data.replace(scrub.encode(), len(scrub) * b'*'))
+        else:
+            log.debug('Sending : %s' % data)
         self.transport.write(data)
 
-    async def execute(self, command: Command) -> Response:
+    async def execute(self, command: Command, scrub: str =None) -> Response:
         if self.state not in Commands.get(command.name).valid_states:
             raise Abort('command %s illegal in state %s' % (command.name, self.state))
 
@@ -422,7 +430,7 @@ class IMAP4ClientProtocol(asyncio.Protocol):
                 await self.pending_async_commands[command.untagged_resp_name].wait()
             self.pending_async_commands[command.untagged_resp_name] = command
 
-        self.send(str(command))
+        self.send(str(command), scrub=scrub)
         try:
             await command.wait()
         except CommandTimeout:
@@ -450,7 +458,9 @@ class IMAP4ClientProtocol(asyncio.Protocol):
     @change_state
     async def login(self, user: str, password: str) -> Response:
         response = await self.execute(
-            Command('LOGIN', self.new_tag(), user, '%s' % quoted(password), loop=self.loop))
+            Command('LOGIN', self.new_tag(), user, '%s' % quoted(password), loop=self.loop),
+            scrub=password,
+        )
 
         if 'OK' == response.result:
             self.state = AUTH
@@ -459,6 +469,27 @@ class IMAP4ClientProtocol(asyncio.Protocol):
                     self.capabilities = self.capabilities.union(set(line.decode().replace('CAPABILITY', '').strip().split()))
         return response
 
+    @change_state
+    async def xoauth2(self, user: str, token: str) -> Response:
+        """Authentication with XOAUTH2.
+
+        Tested with outlook.
+
+        Specification:
+        https://learn.microsoft.com/en-us/exchange/client-developer/legacy-protocols/how-to-authenticate-an-imap-pop-smtp-application-by-using-oauth
+        https://developers.google.com/gmail/imap/xoauth2-protocol
+        """
+        sasl_string = b64encode(f"user={user}\1auth=Bearer {token}\1\1".encode("ascii"))
+
+        response = await self.execute(
+            Command('AUTHENTICATE', self.new_tag(), 'XOAUTH2', sasl_string.decode("ascii"), loop=self.loop),
+            scrub=token,
+        )
+
+        if 'OK' == response.result:
+            self.state = AUTH
+        return response
+
     @change_state
     async def logout(self) -> Response:
         response = (await self.execute(Command('LOGOUT', self.new_tag(), loop=self.loop)))
@@ -668,8 +699,9 @@ class IMAP4ClientProtocol(asyncio.Protocol):
         return [c for c in self.pending_async_commands.values() if c is not None and c.tag == tag]
 
 
-class IMAP4(object):
+class IMAP4:
     TIMEOUT_SECONDS = 10.0
+    _client_task: Task
 
     def __init__(self, host: str = '127.0.0.1', port: int = IMAP4_PORT, loop: asyncio.AbstractEventLoop = None,
                  timeout: float = TIMEOUT_SECONDS, conn_lost_cb: Callable[[Optional[Exception]], None] = None,
@@ -679,13 +711,14 @@ class IMAP4(object):
         self.host = host
         self.protocol = None
         self._idle_waiter = None
+        self.tasks: set[Future] = set()
         self.create_client(host, port, loop, conn_lost_cb, ssl_context)
 
     def create_client(self, host: str, port: int, loop: asyncio.AbstractEventLoop,
                       conn_lost_cb: Callable[[Optional[Exception]], None] = None, ssl_context: ssl.SSLContext = None) -> None:
         local_loop = loop if loop is not None else get_running_loop()
         self.protocol = IMAP4ClientProtocol(local_loop, conn_lost_cb)
-        local_loop.create_task(local_loop.create_connection(lambda: self.protocol, host, port, ssl=ssl_context))
+        self._client_task = local_loop.create_task(local_loop.create_connection(lambda: self.protocol, host, port, ssl=ssl_context))
 
     def get_state(self) -> str:
         return self.protocol.state
@@ -696,6 +729,9 @@ class IMAP4(object):
     async def login(self, user: str, password: str) -> Response:
         return await asyncio.wait_for(self.protocol.login(user, password), self.timeout)
 
+    async def xoauth2(self, user: str, token: bytes) -> Response:
+        return await asyncio.wait_for(self.protocol.xoauth2(user, token), self.timeout)
+
     async def logout(self) -> Response:
         return await asyncio.wait_for(self.protocol.logout(), self.timeout)
 
@@ -744,13 +780,22 @@ class IMAP4(object):
         if self._idle_waiter is not None:
             self._idle_waiter.cancel()
         idle = asyncio.ensure_future(self.idle())
+        self.tasks.add(idle)
+        idle.add_done_callback(self.tasks.discard)
         wait_for_ack = asyncio.ensure_future(self.protocol.wait_for_idle_response())
+        self.tasks.add(wait_for_ack)
+        wait_for_ack.add_done_callback(self.tasks.discard)
         await asyncio.wait({idle, wait_for_ack}, return_when=asyncio.FIRST_COMPLETED)
         if not self.has_pending_idle():
             wait_for_ack.cancel()
             raise Abort('server returned error to IDLE command')
 
-        self._idle_waiter = self.protocol.loop.call_later(timeout, lambda: asyncio.ensure_future(self.stop_wait_server_push()))
+        def start_stop_wait_server_push():
+            task = asyncio.ensure_future(self.stop_wait_server_push())
+            self.tasks.add(task)
+            task.add_done_callback(self.tasks.discard)
+
+        self._idle_waiter = self.protocol.loop.call_later(timeout, start_stop_wait_server_push)
         return idle
 
     def has_pending_idle(self) -> bool:
